<?php

namespace QUI\MasterLogin;

use QUI;
use QUI\Users\User;
use QUI\Utils\Security\Orthos;

/**
 * Class Manager
 *
 * Main manager class for quiqqer/masterlogin
 */
class Manager
{
    /**
     * Tables
     */
    const TBL_LOG = 'masterlogin_log';

    /**
     * Permissions
     */
    const PERMISSION_CAN_USE = 'quiqqer.masterlogin.can_use';
    const PERMISSION_SLAVES  = 'quiqqer.masterlogin.slaves';

    /**
     * Session variables
     */
    const SESSION_MASTER_USER_ID = 'session_master_user_id';
    const SESSION_LOG_ID         = 'session_log_id';

    /**
     * Get IDs of all users that $User is allowed to log in as
     *
     * @param User $User (optional) - [default: current session user]
     * @return int[] - List of User IDs
     */
    public static function getLoginUserIds(User $User = null)
    {
        if (empty($User)) {
            $User = QUI::getUserBySession();
        }

        $Users = QUI::getUsers();

        if ($User->isSU()) {
            $userIds = \array_column($Users->getAllUserIds(), 'id');
        } else {
            $permissionData = $User->getPermission(self::PERMISSION_SLAVES);

            if (empty($permissionData)) {
                return [];
            }

            $slaves = \explode(',', $permissionData);

            $userIds = \array_filter($slaves, function ($entry) {
                return \mb_strpos($entry, 'u') === 0;
            });

            $groupIds = \array_filter($slaves, function ($entry) {
                return \mb_strpos($entry, 'g') === 0;
            });

            // Remove entity identification letter
            $removeIdLetter = function ($value) {
                return (int)\mb_substr($value, 1);
            };

            $userIds  = \array_map($removeIdLetter, $userIds);
            $groupIds = \array_map($removeIdLetter, $groupIds);

            // Parse user IDs from group(s)
            $Groups = QUI::getGroups();

            foreach ($groupIds as $groupId) {
                try {
                    $Group   = $Groups->get($groupId);
                    $userIds = \array_merge(
                        $userIds,
                        $Group->getUserIds()
                    );
                } catch (\Exception $Exception) {
                    QUI\System\Log::writeException($Exception);
                }
            }
        }

        $k = \array_search($User->getId(), $userIds);

        if ($k) {
            unset($userIds[$k]);
        }

        return \array_values(\array_unique($userIds));
    }

    /**
     * Search list of users that $MasterUser is allowed to log in as
     *
     * @param array $searchParams - Search parameters
     * @param User $MasterUser (optional) - [default: current session user]
     * @return int[]|int - List of User IDs or count
     */
    public static function searchLoginUsers($searchParams, User $MasterUser = null)
    {
        $userIds        = [];
        $Grid           = new QUI\Utils\Grid($searchParams);
        $gridParams     = $Grid->parseDBParams($searchParams);
        $usersTbl       = QUI::getDBTableName('users');
        $addressTbl     = QUI::getDBTableName('users_address');
        $binds          = [];
        $where          = [];
        $count          = !empty($searchParams['count']);
        $allowedUserIds = self::getLoginUserIds($MasterUser);

        if ($count) {
            $sql = "SELECT COUNT(*)";
        } else {
            $sql = "SELECT u.`id`, u.`username`, u.`firstname` as user_firstname, u.`lastname` as user_lastname, "
                   ."a.`firstname` as address_firstname, a.`lastname` as address_lastname, u.`email`, a.`company`";
        }

        $sql     .= " FROM `".$usersTbl."` as u LEFT JOIN `".$addressTbl."` as a ON u.`id` = a.`uid`";
        $where[] = 'u.`active` = 1';

        if (!empty($allowedUserIds)) {
            $where[] = 'u.`id` IN ('.\implode(',', $allowedUserIds).')';
        }

        if (!empty($searchParams['search'])) {
            $whereOR = [];

            $searchColumns = [
                'u.`id`',
                'u.`username`',
                'u.`firstname`',
                'u.`lastname`',
                'a.`firstname`',
                'a.`lastname`',
                'u.`email`',
                'a.`company`',
                'a.`city`'
            ];

            foreach ($searchColumns as $tbl => $column) {
                $whereOR[]       = $column.' LIKE :search';
                $binds['search'] = [
                    'value' => '%'.$searchParams['search'].'%',
                    'type'  => \PDO::PARAM_STR
                ];
            }

            $where[] = '('.implode(' OR ', $whereOR).')';
        }

        // build WHERE query string
        if (!empty($where)) {
            $sql .= " WHERE ".implode(" AND ", $where);
        }

        // ORDER
        if (!empty($searchParams['sortOn'])) {
            if (!empty($searchParams['sortBy'])) {
                $sortDirection = Orthos::clear($searchParams['sortBy']);
            } else {
                $sortDirection = "ASC";
            }

            switch (Orthos::clear($searchParams['sortOn'])) {
                case 'name':
                    $sortColumns = [
                        'u.`username`',
                        'u.`firstname`',
                        'u.`lastname`',
                        'a.`firstname`',
                        'a.`lastname`'
                    ];
                    break;

                case 'email':
                    $sortColumns = [
                        'u.`email`'
                    ];
                    break;

                case 'company':
                    $sortColumns = [
                        'a.`company`'
                    ];
                    break;

                default:
                    $sortColumns = [
                        'u.`id`',
                    ];
            }

            $order = [];

            foreach ($sortColumns as $column) {
                $order[] = "$column $sortDirection";
            }

            $sql .= " ORDER BY ".\implode(',', $order);
        }

        // LIMIT
        if (!empty($gridParams['limit']) && !$count) {
            $sql .= " LIMIT ".$gridParams['limit'];
        } else {
            if (!$count) {
                $sql .= " LIMIT ".(int)20;
            }
        }

        $Stmt = QUI::getPDO()->prepare($sql);

        // bind search values
        foreach ($binds as $var => $bind) {
            $Stmt->bindValue(':'.$var, $bind['value'], $bind['type']);
        }

        try {
            $Stmt->execute();
            $result = $Stmt->fetchAll(\PDO::FETCH_ASSOC);
        } catch (\Exception $Exception) {
            QUI\System\Log::addError(
                self::class.' :: searchUsers() -> '.$Exception->getMessage()
            );

            return [];
        }

        if ($count) {
            return (int)current(current($result));
        }

        $users = [];

        foreach ($result as $row) {
            $userId = (int)$row['id'];

            if (isset($users[$userId])) {
                continue;
            }

            // Name
            $name = $row['user_firstname'].' '.$row['user_lastname'];

            if (!empty($row['address_firstname']) && !empty($row['address_lastname'])) {
                $name = $row['address_firstname'].' '.$row['address_lastname'];
            }

            $users[$userId] = [
                'id'      => $userId,
                'name'    => $name,
                'email'   => $row['email'],
                'company' => $row['company']
            ];
        }

        return \array_values($users);
    }

    /**
     * Switches the current session user to $slaveUserId (requires permission!)
     *
     * @param int $slaveUserId
     * @return void
     *
     * @throws PermissionDeniedException
     */
    public static function logInAs(int $slaveUserId)
    {
        if (self::isLoggedInAsSlave()) {
            throw new PermissionDeniedException();
        }

        if (!\in_array($slaveUserId, self::getLoginUserIds())) {
            throw new PermissionDeniedException();
        }

        $masterUserId = QUI::getUserBySession()->getId();
        $Session      = QUI::getSession();
        $Session->set(self::SESSION_MASTER_USER_ID, $masterUserId);
        $Session->set('uid', $slaveUserId);

        // Create log entry
        try {
            QUI::getDataBase()->insert(
                self::getLogTable(),
                [
                    'master_id' => $masterUserId,
                    'slave_id'  => $slaveUserId,
                    'start'     => \date('Y-m-d H:i:s'),
                    'sec_hash'  => QUI::getUsers()->getSecHash()
                ]
            );

            $Session->set(self::SESSION_LOG_ID, QUI::getDataBase()->getPDO()->lastInsertId());
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }
    }

    /**
     * Switches the current session user to the original user
     *
     * @return void
     */
    public static function logBack()
    {
        if (!self::isLoggedInAsSlave()) {
            return;
        }

        $Session      = QUI::getSession();
        $masterUserId = $Session->get(self::SESSION_MASTER_USER_ID);

        $Session->set('uid', $masterUserId);
        $Session->set(self::SESSION_MASTER_USER_ID, false);

        // Update log entry
        $logId = $Session->get(self::SESSION_LOG_ID);

        if (empty($logId)) {
            return;
        }

        try {
            QUI::getDataBase()->update(
                self::getLogTable(),
                [
                    'end' => \date('Y-m-d H:i:s')
                ],
                [
                    'id' => $logId
                ]
            );

            $Session->set(self::SESSION_LOG_ID, false);
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }
    }

    /**
     * Checks if the current Session user is logged in as a slave user
     *
     * @return bool
     */
    public static function isLoggedInAsSlave()
    {
        $masterUserId = QUI::getSession()->get(self::SESSION_MASTER_USER_ID);
        return !empty($masterUserId);
    }

    /**
     * Check if $User is allowed to use the Master Login feature and throws and exception if not
     *
     * @param User $User (optional) - [default: session user]
     * @return void
     *
     * @throws PermissionDeniedException
     */
    public static function checkUsePermission(User $User = null)
    {
        if (!self::canUse($User)) {
            throw new PermissionDeniedException();
        }
    }

    /**
     * Check if $User is allowed to use the Master Login feature
     *
     * @param User $User (optional) - [default: session user]
     * @return bool
     */
    public static function canUse(User $User = null)
    {
        if (self::isLoggedInAsSlave()) {
            return true;
        }

        if (empty($User)) {
            $User = QUI::getUserBySession();
        }

        if ($User->isSU()) {
            return true;
        }

        return $User->hasPermission(self::PERMISSION_CAN_USE);
    }

    /**
     * Get log table name
     *
     * @return string
     */
    public static function getLogTable()
    {
        return QUI::getDBTableName(self::TBL_LOG);
    }
}
